import parser._
import util.Random

package object microc {

  trait Exp
  // Aritmética
  case class Num(v: Double) extends Exp
  case class Soma(e1: Exp, e2: Exp) extends Exp
  case class Mult(e1: Exp, e2: Exp) extends Exp
  // Condicionais
  case class Menor(e1: Exp, e2: Exp) extends Exp
  case class If(econd: Exp, ethen: Exp, eelse: Exp) extends Exp
  // Funções de primeira ordem
  case class Fun1(nome: String, params: List[String], corpo: Exp)
  case class Prog(funs: Set[Fun1], corpo: Exp)
  case class Var(nome: String) extends Exp
  case class Ap1(fun: String, args: List[Exp]) extends Exp
  // Let
  case class Let(nome: String, exp: Exp, corpo: Exp) extends Exp
  // Referências
  case class Seq(e1: Exp, e2: Exp) extends Exp
  case class Atrib(lval: Exp, rval: Exp) extends Exp
  case class Ref(e: Exp) extends Exp
  case class Deref(e: Exp) extends Exp

  type Valor = Int
  type End = Int
  type Mem = Map[End, Valor]

  // Ações
  type Acao[T] = Mem => (T, Mem)

  implicit class RichAcao[A](val p: Acao[A]) extends AnyVal {
    def flatMap[B](f: A => Acao[B]): Acao[B] = bind(p, f)
    def map[B](f: A => B): Acao[B] = bind(p, (v: A) => emptya(f(v)))
  }

  def emptya[T](v: T): Acao[T] = m => (v, m)
  def le(r: Int): Acao[Valor] = m => (m(r), m)
  def escreve(r: Int, v: Valor): Acao[Valor] = m => (r, m + (r -> v))
  def bind[T, U](a: Acao[T], f: T => Acao[U]): Acao[U] = m => {
    val (v, nm) = a(m)
    f(v)(nm)
  }

  def aloca(n: Valor): Acao[End] = for {
    l <- le(0)
    _ <- escreve(l, n)
    _ <- escreve(0, l+1)
  } yield l

  def lift[T](l: List[Acao[T]]): Acao[List[T]] = l match {
    case Nil => emptya(Nil)
    case h :: t => for {
      vh <- h // h é uma Acao[T]
      vt <- lift(t) // lift(t) é uma Acao[List[T]]
    } yield vh :: vt
  }

  def ProgFun: Parser[Prog] = for {
    funs <- many(DefFun)
    corpo <- ExpFun
    _ <- space
    _ <- not(pred(c => true), ())
  } yield Prog(funs.toSet, corpo)

  def DefFun: Parser[Fun1] = for {
    _ <- kw("fun")
    (nome, _) <- id
    _ <- op("(")
    params <- (for {
      (p, _) <- id
      ps <- many(for {
        _ <- op(",")
        (p, _) <- id
      } yield p)
    } yield p :: ps) +: empty(List())
    _ <- op(")")
    corpo <- ExpFun
    _ <- kw("end")
  } yield Fun1(nome, params, corpo)

  def ExpFun: Parser[Exp] =
    chainl(AtexpFun,
      for {_ <- op(";")} yield (e1, e2) => Seq(e1, e2),
      AtexpFun)

  def AtexpFun: Parser[Exp] =
    chainl(RexpFun,
      for {_ <- op(":=")} yield (e1, e2) => Atrib(e1, e2),
      RexpFun)

  def RexpFun: Parser[Exp] =
    chainl(AexpFun,
      for {_ <- op("<")} yield (e1, e2) => Menor(e1, e2),
      AexpFun)

  def AexpFun: Parser[Exp] =
    chainl(TermoFun,
      (for {_ <- op("+")} yield (e1, e2) => Soma(e1, e2)) +:
        (for {_ <- op("-")} yield (e1: Exp, e2: Exp) =>
          Soma(e1, Mult(Num(-1), e2))),
      TermoFun)

  def TermoFun: Parser[Exp] =
    chainl(FatorFun, for {
      _ <- op("*")
    } yield (e1, e2) => Mult(e1, e2), FatorFun)

  def ArgsFun: Parser[List[Exp]] = (for {
    a <- ExpFun
    as <- many(for {
      _ <- op(",")
      a <- ExpFun
    } yield a)
  } yield a :: as) +: empty(List())

  def FatorFun: Parser[Exp] =
    (for {(v, pos) <- num} yield Num(v)) +:
      (for {
        _ <- op("(")
        e <- ExpFun
        _ <- op(")")
      } yield e) +:
      (for {
        _ <- op("-")
        e <- FatorFun
      } yield Mult(Num(-1), e)) +:
      (for {
        _ <- op("&")
        e <- FatorFun
      } yield Ref(e)) +:
      /*(for {
        _ <- kw("throw")
        (msg, _) <- str
      } yield Throw(msg)) +:*/
      (for {
        _ <- op("*")
        e <- FatorFun
      } yield Deref(e)) +:
      (for {
        _ <- kw("if")
        econd <- ExpFun
        _ <- kw("then")
        ethen <- ExpFun
        _ <- kw("else")
        eelse <- ExpFun
        _ <- kw("end")
      } yield If(econd, ethen, eelse)) +:
      /*(for {
        _ <- kw("try")
        etry <- ExpFun
        _ <- kw("catch")
        ecatch <- ExpFun
        _ <- kw("end")
      } yield TryCatch(etry, ecatch)) +:
      (for {
        _ <- kw("try")
        etry <- ExpFun
        _ <- kw("finally")
        efin <- ExpFun
        _ <- kw("end")
      } yield TryFinally(etry, efin)) +:
      (for {
        _ <- kw("try")
        etry <- ExpFun
        _ <- kw("catch")
        ecatch <- ExpFun
        _ <- kw("finally")
        efin <- ExpFun
        _ <- kw("end")
      } yield TryFinally(TryCatch(etry, ecatch), efin)) +:*/
      (for {
        _ <- kw("let")
        (nome, _) <- id
        _ <- op("=")
        exp <- ExpFun
        resto <- many(for {
          _ <- op(",")
          (nome, _) <- id
          _ <- op("=")
          exp <- ExpFun
        } yield (nome, exp))
        _ <- kw("in")
        corpo <- ExpFun
        _ <- kw("end")
      } yield ((nome, exp) :: resto).foldRight(corpo) {
        case ((nome, exp), corpo) => Let(nome, exp, corpo) // Ap(Fun(List(nome), corpo), List(exp))
      }) +:
      (for {
        (nome, _) <- id
        _ <- op("(")
        args <- (for {
          a <- ExpFun
          as <- many(for {
            _ <- op(",")
            a <- ExpFun
          } yield a)
        } yield a :: as) +: empty(List())
        _ <- op(")")
      } yield Ap1(nome, args)) +:
      (for {(nome, _) <- id} yield Var(nome))

  def eval(p: Prog): (Valor, Mem) = p match {
    case Prog(funs, corpo) => eval(funs)(corpo)(Map(0 -> 1))
  }

  def eval(funs: Set[Fun1]) = {
    def eval(e: Exp): Acao[Valor] = e match {
      case Num(v) => emptya(v.toInt)
      case Soma(e1, e2) => for {
        n1 <- eval(e1)
        n2 <- eval(e2)
      } yield n1 + n2
      case Mult(e1, e2) => for {
        n1 <- eval(e1)
        n2 <- eval(e2)
      } yield n1 * n2
      case Menor(e1, e2) => for {
        n1 <- eval(e1)
        n2 <- eval(e2)
      } yield if(n1 < n2) 1 else 0
      case If(ec, et, ee) => for {
        vc <- eval(ec)
        vif <- if(vc != 0) eval(et) else eval(ee)
      } yield vif
      case Ap1(nome, args) =>
        val Fun1(_, params, corpo) = funs.find(f => f.nome == nome).get
        for {
          vas <- lift(args.map(a => eval(a)))  // vas é uma List[Valor]
          ves <- lift(vas.map(n => aloca(n)))
          vap <- eval(subst(params.zip(ves.map(n => Deref(Num(n)))).toMap)(corpo))
          sp <- le(0)
          _ <- escreve(0, sp - params.size)
        } yield vap
      case Seq(e1, e2) => for {
        _ <- eval(e1)
        ve2 <- eval(e2)
      } yield ve2
      case Atrib(Deref(lv), rv) => for {
        l <- eval(lv)
        vrv <- eval(rv)
        c <- escreve(l, vrv)
      } yield c
      case Deref(e) => for {
        l <- eval(e)
        v <- le(l)
      } yield v
      case Ref(Deref(lv)) => for {
        l <- eval(lv)
      } yield l
      case Let(n, e, c) => for {
        ve <- eval(e)
        l <- aloca(ve)
        vc <- eval(subst(Map(n -> Deref(Num(l))))(c))
        sp <- le(0)
        _ <- escreve(0, sp - 1)
      } yield vc
    }
    eval _ // retorna a função eval interna (com funs fixado)
  }

  def fvs(e: Exp): Set[String] = e match {
    case Soma(e1, e2) => fvs(e1) ++ fvs(e2)
    case Mult(e1, e2) => fvs(e1) ++ fvs(e2)
    case Menor(e1, e2) => fvs(e1) ++ fvs(e2)
    case If(ec, et, ee) => fvs(ec) ++ fvs(et) ++ fvs(ee)
    case Var(x) => Set(x)
    case Ap1(nome, args) => args.map(arg => fvs(arg)).reduceOption((a, b) => a ++ b).getOrElse(Set())
    case Let(nome, exp, corpo) => fvs(exp) ++ (fvs(corpo) - nome)
    case Seq(e1, e2) => fvs(e1) ++ fvs(e2)
    case Atrib(lv, rv) => fvs(lv) ++ fvs(rv)
    case Ref(e) => fvs(e)
    case Deref(e) => fvs(e)
    /*case TryCatch(e1, e2) => fvs(e1) ++ fvs(e2)
    case TryFinally(e1, e2) => fvs(e1) ++ fvs(e2)*/
    case _ => Set()
  }

  def subst(n2v: Map[String, Exp]): (Exp => Exp) = {
    def sub(e: Exp): Exp = e match {
      case Soma(e1, e2) => Soma(sub(e1), sub(e2))
      case Mult(e1, e2) => Mult(sub(e1), sub(e2))
      case Menor(e1, e2) => Menor(sub(e1), sub(e2))
      case If(ec, et, ee) => If(sub(ec), sub(et), sub(ee))
      case Ap1(nome, args) => Ap1(nome, args.map(arg => sub(arg)))
      case Var(nome) => n2v.get(nome) match {
        case Some(e) => e
        case None => Var(nome)
      }
      case Let(nome, exp, corpo) =>
        val scorpo = n2v - nome
        // variáveis no mapa de substituição que são efetivamente usadas
        val conflito = fvs(corpo).intersect(scorpo.keySet)
        // variáveis livres nos valores do mapa de substituição
        val freevars = scorpo.filterKeys(n => conflito.contains(n)).values.map(e => fvs(e)).reduceOption((a, b) => a ++ b).getOrElse(Set())
        if(freevars.contains(nome)) { // ocorreu uma captura
        // Renomeia nome para evitar colisão com seu uso em u
        val novonome = Random.alphanumeric.take(10).mkString
          Let(novonome, sub(exp), subst(scorpo)(subst(Map(nome -> Var(novonome)))(corpo)))
        } else Let(nome, sub(exp), subst(scorpo)(corpo))
      case Seq(e1, e2) => Seq(sub(e1), sub(e2))
      case Atrib(lv, rv) => Atrib(sub(lv), sub(rv))
      case Ref(e) => Ref(sub(e))
      case Deref(e) => Deref(sub(e))
      /*case TryCatch(et, ec) => TryCatch(sub(et), sub(ec))
      case TryFinally(et, ef) => TryFinally(sub(et), sub(ef))*/
      case _ => e
    }
    sub
  }

}
